/*
 * @author     Ekawat (ice) Homsirikamol
 * @version    1.0
 * @copyright  Copyright (c) 2017 Cryptographic Engineering Research Group
 *             ECE Department, George Mason University Fairfax, VA, U.S.A.
 *             All rights Reserved.
 * @license    This project is released under the GNU Public License.
 *             The license and distribution terms for this file may be
 *             found in the file LICENSE in this distribution or at
 *             http://www.gnu.org/licenses/gpl-3.0.txt
 * @note       This implementation is derived in part from the reference implementation
 *             included in SUPERCOP-20171020, with the list of co-authors listed in ref/designers
 *             
 *             This is publicly available encryption source code that falls
 *             under the License Exception TSU (Technology and software-
 *             unrestricted)
 */

#if !__SYNTHESIS__
#include <iostream>
#endif

#include "ap_int.h"
#include "hls_stream.h"
#include "hls_interface.h"
using namespace hls;

#if !__SYNTHESIS__
#define DBG 1
#define DBG_INPUT 1
#endif

#define FAIL_SYN 0


#define INIT_ROUNDS 16
#define TAG_ROUNDS  10
#define MW 64   // MORUS word size
#define n1 13
#define n2 46
#define n3 38
#define n4 7
#define n5 4
#define rotl(x,n)      (((x) << (n)) | ((x) >> (64-n)))

enum fsms_t {S_INIT, S_INIT_FINAL, 
    S_DATA, 
    S_TAG_INIT, 
    S_TAGX};

/* ------------------------------------------------------------------------- */
ap_uint<32> swap32(ap_uint<32> x)
{
#pragma HLS INLINE
    x = (x & 0x0000FFFF) << 16 | (x & 0xFFFF0000) >> 16;
    x = (x & 0x00FF00FF) << 8  | (x & 0xFF00FF00) >> 8;
    return x;
}
ap_uint<64> swap64(ap_uint<64> x)
{
#pragma HLS INLINE
    x = (x & 0x00000000FFFFFFFF) << 32 | (x & 0xFFFFFFFF00000000) >> 32;
    x = (x & 0x0000FFFF0000FFFF) << 16 | (x & 0xFFFF0000FFFF0000) >> 16;
    x = (x & 0x00FF00FF00FF00FF) << 8  | (x & 0xFF00FF00FF00FF00) >> 8;
    return x;
}

static void reshape_block64(array_t word, data_t block)
{
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=word           complete dim=1
    int i;
    for (i=0; i<ELEMENTS; i++) {
    #pragma HLS UNROLL
        ap_uint<64> tmp = (ap_uint<ELEMENTSIZE>) (block >> ((ELEMENTS-i-1)*ELEMENTSIZE));
        word[i] = swap64(tmp);
    }
}

static data_t reshape_array64(array_t word)
{
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=word           complete dim=1
    int i;
    data_t tmp = 0;
    for (i=0; i<ELEMENTS; i++) {
    #pragma HLS UNROLL
        ap_uint<64> t = swap64(word[i]);
        tmp ^= ((data_t) t) << ((ELEMENTS-i-1)*ELEMENTSIZE);
    }
    return tmp;
}

static void clear(array_t a) 
{
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=a complete dim=1
    int i;
    for(i=0; i<ELEMENTS; i++) {
    #pragma HLS UNROLL
        a[i] = 0;
    }
}

static void copy(array_t b, array_t a) 
{
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=a complete dim=1
#pragma HLS ARRAY_RESHAPE variable=b complete dim=1
    int i;
    for(i=0; i<ELEMENTS; i++) {
    #pragma HLS UNROLL
        b[i] = a[i];
    }
}
/* ------------------------------------------------------------------------- */

void morus_stateupdate(ap_uint<MW> msgblk[4], ap_uint<MW> state[5][4])
{
    ap_uint<64> temp, temp1;
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=msgblk  complete dim=1
#pragma HLS ARRAY_RESHAPE variable=state complete dim=1
#pragma HLS ARRAY_RESHAPE variable=state[0] complete dim=1
#pragma HLS ARRAY_RESHAPE variable=state[1] complete dim=1
#pragma HLS ARRAY_RESHAPE variable=state[2] complete dim=1
#pragma HLS ARRAY_RESHAPE variable=state[3] complete dim=1
#pragma HLS ARRAY_RESHAPE variable=state[4] complete dim=1

	state[0][0] ^= state[3][0]; state[0][1] ^= state[3][1]; state[0][2] ^= state[3][2]; state[0][3] ^= state[3][3];
	temp = state[3][3];    state[3][3] = state[3][2];  state[3][2] = state[3][1];  state[3][1] = state[3][0];  state[3][0] = temp;
	state[0][0] ^= state[1][0] & state[2][0]; state[0][1] ^= state[1][1] & state[2][1]; state[0][2] ^= state[1][2] & state[2][2]; state[0][3] ^= state[1][3] & state[2][3];
	state[0][0] = rotl(state[0][0], n1);  state[0][1] = rotl(state[0][1], n1);       state[0][2] = rotl(state[0][2], n1);       state[0][3] = rotl(state[0][3], n1);

	state[1][0] ^= msgblk[0];   state[1][1] ^= msgblk[1];   state[1][2] ^= msgblk[2];   state[1][3] ^= msgblk[3];
	state[1][0] ^= state[4][0]; state[1][1] ^= state[4][1]; state[1][2] ^= state[4][2]; state[1][3] ^= state[4][3];
	temp = state[4][3];    state[4][3] = state[4][1];  state[4][1] = temp;
	temp1 = state[4][2];    state[4][2] = state[4][0];  state[4][0] = temp1;
	state[1][0] ^= (state[2][0] & state[3][0]); state[1][1] ^= (state[2][1] & state[3][1]); state[1][2] ^= (state[2][2] & state[3][2]); state[1][3] ^= (state[2][3] & state[3][3]);
	state[1][0] = rotl(state[1][0], n2);  state[1][1] = rotl(state[1][1], n2);       state[1][2] = rotl(state[1][2], n2);       state[1][3] = rotl(state[1][3], n2);

	state[2][0] ^= msgblk[0];   state[2][1] ^= msgblk[1];   state[2][2] ^= msgblk[2];   state[2][3] ^= msgblk[3];
	state[2][0] ^= state[0][0]; state[2][1] ^= state[0][1]; state[2][2] ^= state[0][2]; state[2][3] ^= state[0][3];
	temp = state[0][0];    state[0][0] = state[0][1];  state[0][1] = state[0][2];  state[0][2] = state[0][3];  state[0][3] = temp;
	state[2][0] ^= state[3][0] & state[4][0]; state[2][1] ^= state[3][1] & state[4][1]; state[2][2] ^= state[3][2] & state[4][2]; state[2][3] ^= state[3][3] & state[4][3];
	state[2][0] = rotl(state[2][0], n3);  state[2][1] = rotl(state[2][1], n3);       state[2][2] = rotl(state[2][2], n3);       state[2][3] = rotl(state[2][3], n3);

	state[3][0] ^= msgblk[0];   state[3][1] ^= msgblk[1];   state[3][2] ^= msgblk[2];   state[3][3] ^= msgblk[3];
	state[3][0] ^= state[1][0]; state[3][1] ^= state[1][1]; state[3][2] ^= state[1][2]; state[3][3] ^= state[1][3];
	temp = state[1][3];    state[1][3] = state[1][1];  state[1][1] = temp;
	temp1 = state[1][2];    state[1][2] = state[1][0];  state[1][0] = temp1;
	state[3][0] ^= state[4][0] & state[0][0]; state[3][1] ^= state[4][1] & state[0][1]; state[3][2] ^= state[4][2] & state[0][2]; state[3][3] ^= state[4][3] & state[0][3];
	state[3][0] = rotl(state[3][0], n4);  state[3][1] = rotl(state[3][1], n4);       state[3][2] = rotl(state[3][2], n4);       state[3][3] = rotl(state[3][3], n4);

	state[4][0] ^= msgblk[0];   state[4][1] ^= msgblk[1];   state[4][2] ^= msgblk[2];   state[4][3] ^= msgblk[3];
	state[4][0] ^= state[2][0]; state[4][1] ^= state[2][1]; state[4][2] ^= state[2][2]; state[4][3] ^= state[2][3];
	temp = state[2][3];    state[2][3] = state[2][2];  state[2][2] = state[2][1];  state[2][1] = state[2][0];  state[2][0] = temp;
	state[4][0] ^= state[0][0] & state[1][0]; state[4][1] ^= state[0][1] & state[1][1]; state[4][2] ^= state[0][2] & state[1][2]; state[4][3] ^= state[0][3] & state[1][3];
	state[4][0] = rotl(state[4][0], n5);  state[4][1] = rotl(state[4][1], n5);       state[4][2] = rotl(state[4][2], n5);       state[4][3] = rotl(state[4][3], n5);
}

/* ------------------------------------------------------------------------- */

void hls_ciphercore (
    BDO_STREAM& bdo_bus,
    bool *msg_auth_valid,
    BDI_STREAM& bdi_bus,
    KEY_STREAM& key_bus,
    bool decrypt,
    volatile bool key_update
    )
{
    /* Top-Level Interface */
#pragma HLS DATA_PACK variable=bdo_bus
#pragma HLS DATA_PACK variable=bdi_bus
#pragma HLS DATA_PACK variable=key_bus
#pragma HLS INTERFACE port=bdo_bus axis
#pragma HLS INTERFACE port=bdi_bus axis
#pragma HLS INTERFACE port=key_bus axis
#pragma HLS INTERFACE ap_vld  port=msg_auth_valid
#if __SYNTHESIS__
#else
    size_t dbg_loop = 0;
#endif

    /* Interface related temporary variables */
    public_bus bdi;
    output_bus bdo;
    secret_bus key;


    fsms_t fsm = S_INIT;
    bool is_eoi_r = 0;
    static ap_uint<MW> key_reg[KEYSIZE/MW];
    ap_uint<4> cnt = 0;
    ap_uint<MW> state[5][4];
    ap_uint<MAX_LEN> len_d = 0;
    ap_uint<MAX_LEN> len_ad = 0;
    const ap_uint<8> con[32] = {0x0,0x1,0x01,0x02,0x03,0x05,0x08,0x0d,0x15,0x22,0x37,0x59,0x90,0xe9,0x79,0x62, 0xdb, 0x3d, 0x18, 0x55, 0x6d, 0xc2, 0x2f, 0xf1, 0x20, 0x11, 0x31, 0x42, 0x73, 0xb5, 0x28, 0xdd};
    array_t tt;
    array_t npub_data;

#pragma HLS ARRAY_RESHAPE variable=key_reg complete dim=1
#pragma HLS ARRAY_RESHAPE variable=state complete dim=1
#pragma HLS ARRAY_RESHAPE variable=state[0] complete dim=1
#pragma HLS ARRAY_RESHAPE variable=state[1] complete dim=1
#pragma HLS ARRAY_RESHAPE variable=state[2] complete dim=1
#pragma HLS ARRAY_RESHAPE variable=state[3] complete dim=1
#pragma HLS ARRAY_RESHAPE variable=state[4] complete dim=1
#pragma HLS ARRAY_RESHAPE variable=con complete dim=1
#pragma HLS ARRAY_RESHAPE variable=tt complete dim=1
#pragma HLS ARRAY_RESHAPE variable=npub_data complete dim=1

    // Key initialization (if needed)
    if (key_update == 1) {
#if KEYBUS == 32
        LOAD_KEY: for (size_t i=0; i<4; i++) {
            key_bus.read(key);
            #if MW == 64
                key_reg[0] = (key_reg[0] >> 32) ^ (key_reg[1] << 32);
                key_reg[1] = (key_reg[1] >> 32) ^ (((ap_uint<64>) swap32(key.data)) << 32);
            #elif MW == 32
                key_reg[0] = key_reg[1];
                key_reg[1] = key_reg[2];
                key_reg[2] = key_reg[3];
                key_reg[3] = swap32(key.data);
            #endif
        }
#endif
    }

    // Initialize here
    bdi_bus.read(bdi);
    reshape_block64(npub_data, bdi.data);
    bool empty = bdi.eoi;

#if MW == 64
    state[0][0] = npub_data[0];
    state[0][1] = npub_data[1];
    state[0][2] = 0;
    state[0][3] = 0;
#endif
    ap_uint<64> tcon;
    for (size_t i=0; i<4; i++) {
    #pragma HLS UNROLL
        state[1][i] = key_reg[i % 2];
        state[2][i] = 0xFFFFFFFFFFFFFFFF;
        state[3][i] = 0;
        tcon = 0;
        for (size_t j=0; j<8; j++) {
        #pragma HLS UNROLL
            tcon = (tcon << 8) ^ con[i*8+j];
        }
        state[4][i] = swap64(tcon);
    }

    bool loop = 1;

    while (loop) {
        // Add this clause to prevent infinite loop during debugging
#if !__SYNTHESIS__
        if (dbg_loop > 100) return;
        else dbg_loop++;
#endif
#if DBG
        std::cout << "Current fsm = " << fsm << " - cnt = " << cnt << std::endl;
        printf("Input state:\n");
        for(size_t j=0; j<5; j++) {
            for(size_t i=0; i<4; i++)
                printf("%016llX  ", (unsigned long long) state[j][i]);
            printf("\n");
        }
#endif
        /* Default values */
        bool is_eoi = 0;
        bool is_data = 0;
        bsize_t size = 0;
        array_t bdi_data;
        array_t o1, output;
        array_t input;
        ap_uint<8> o2[BLOCKSIZE/8];
        ap_uint<8> o3[BLOCKSIZE/8];
        ap_uint<BLOCKSIZE/8> vbytes = 0;
       
        /* ==================== */
        /* Preparing input data */
        /* ==================== */
        if (fsm == S_DATA) {
            bdi_bus.read(bdi); 
            reshape_block64(bdi_data, bdi.data);

            is_data = (bdi.type == TYPE_D) ? 1:0;
            is_eoi = bdi.eoi;
            size = bdi.size;
            vbytes = bdi.valid_bytes;
            #if DBG
               std::cout << "Reading ..." << std::endl;
            #endif
        } else {
            clear(bdi_data);
        }

        for (size_t i=0; i<4; i++) {
        #pragma HLS UNROLL
            o1[i] = bdi_data[i] ^ state[0][i] ^ state[1][(i+1) % 4] ^ (state[2][i] & state[3][i]);
            // Select only the valid bytes
            for (size_t j=0; j<8; j++) {
            #pragma HLS UNROLL
                o2[i*8+j] = o1[i] >> (j*8);
                if ((vbytes >> BLOCKSIZE/8-1-(i*8+j)) & 0x1)
                    o3[i*8+j] = o2[i*8+j];
                else
                    o3[i*8+j] = 0;
            }
            ap_uint<64> tmp = 0;
            for (size_t j=0; j<8; j++) {
            #pragma HLS UNROLL
                tmp = (tmp << 8) ^ o3[i*8+j];
            }
            output[i] = swap64(tmp);
        }

        if (is_data) {
            bdo.data = reshape_array64(output);
            bdo_bus.write(bdo);
            #if DBG
               std::cout << "Writing ..." << std::endl;
            #endif
        }
        #if DBG_INPUT_OLD
            printf("BDI data:\n");
            for(size_t i=0; i<4; i++)
                printf("%016llX  ", (unsigned long long) bdi_data[i]);
            printf("\n");
            printf("O1 data:\n");
            for(size_t i=0; i<4; i++)
                printf("%016llX  ", (unsigned long long) o1[i]);
            printf("\n");
            printf("O2 data:\n");
            for(size_t i=0; i<4; i++) {
                for (size_t j=0; j<8; j++)
                    printf("%02X", (unsigned char) o2[i*8+j]);
                printf(" ");
            }
            printf("\n");
            printf("VBYTES: %08X\n", (unsigned int) vbytes);
            printf("O3 data:\n");
            for(size_t i=0; i<4; i++) {
                for (size_t j=0; j<8; j++)
                    printf("%02X", (unsigned char) o3[i*8+j]);
                printf(" ");
            }
            printf("\n");
            printf("Out data:\n");
            for(size_t i=0; i<4; i++)
                printf("%016llX  ", (unsigned long long) output[i]);
            printf("\n");
        #endif

        if (fsm == S_TAGX)
            copy(input, tt);
        else if (fsm == S_DATA && (is_data && decrypt))
            copy(input, output);
        else
            copy(input, bdi_data);
        
        if (fsm == S_TAG_INIT) {
            tt[0] = (len_ad << 3);
            tt[1] = (len_d << 3);
            tt[2] = 0;
            tt[3] = 0;
        }
        /* ========= */
        /* MAIN OP   */
        /* ========= */
        #if DBG_INPUT_OLD
            printf("Real Input state:\n");
            for(size_t j=0; j<5; j++) {
                for(size_t i=0; i<4; i++)
                    printf("%016llX  ", (unsigned long long) state[j][i]);
                printf("\n");
            }
        #endif

        if (fsm == S_INIT_FINAL) {
            state[1][0] ^= key_reg[0];
            state[1][1] ^= key_reg[1];
            state[1][2] ^= key_reg[0];
            state[1][3] ^= key_reg[1];
        } else if ( fsm == S_TAG_INIT) {
            ap_uint<MW> to_state[5][4];

            state[4][0] ^= state[0][0]; 
            state[4][1] ^= state[0][1]; 
            state[4][2] ^= state[0][2]; 
            state[4][3] ^= state[0][3];
        } else {
            #if DBG_INPUT
                printf("Input data:\n");
                for(size_t i=0; i<4; i++)
                    printf("%016llX  ", (unsigned long long) input[i]);
                printf("\n");
            #endif
            morus_stateupdate(input, state);
        }
        
        #if DBG_Output
            printf("Output state:\n");
            for(size_t j=0; j<5; j++) {
                for(size_t i=0; i<4; i++)
                    printf("%016llX  ", (unsigned long long) state[j][i]);
                printf("\n");
            }
        #endif
        /* ========= */
        /* Post processing  */
        /* ========= */

#if FAIL_SYN
        if (fsm == S_INIT) {
            if (cnt == INIT_ROUNDS-1) {
                cnt = 0; 
                fsm = S_INIT_FINAL;
            } else {
                cnt++;
            }
        } else if (fsm == S_INIT_FINAL) {
            if (empty) fsm = S_TAG_INIT;
            else fsm = S_DATA;
            #if DBG
                if (empty) std::cout << "\n Tag \n" << std::endl;
                else std::cout << "\n Data \n" << std::endl;
            #endif
        } else if (fsm == S_DATA) {
            if (is_data) len_d += size;
            else len_ad += size;
            if (is_eoi) fsm = S_TAG_INIT;
            #if DBG
                if (is_eoi) std::cout << "\n Tag \n" << std::endl;
            #endif
        } else if (fsm == S_TAG_INIT) {
            fsm = S_TAGX;
        } else if (fsm == S_TAGX) {
            if (cnt == TAG_ROUNDS-1) loop = 0;
            else cnt++;
        }
#else
        switch (fsm) {
        case S_INIT:
            if (cnt == INIT_ROUNDS-1) { 
                cnt = 0;
                fsm = S_INIT_FINAL;
            } else { 
                cnt++;
            }
            break;
        case S_INIT_FINAL:
            if (empty) fsm = S_TAG_INIT;
            else fsm = S_DATA;
            #if DBG
                if (empty) std::cout << "\n Tag \n" << std::endl;
                else std::cout << "\n Data \n" << std::endl;
            #endif
            break;
        case S_DATA:
            if (is_data) len_d += size;
            else len_ad += size;
            if (is_eoi) fsm = S_TAG_INIT;
            #if DBG
                if (is_eoi) std::cout << "\n Tag \n" << std::endl;
            #endif
            break;
        case S_TAG_INIT:
            fsm = S_TAGX;
            break;
        default:
            if (cnt == TAG_ROUNDS-1) loop = 0;
            else cnt++;
        }
    }
#endif

    // Put this outside loop so it doesn't take up clock cycle in the main loop
#if DBG
    printf("Final state:\n");
    for(size_t j=0; j<5; j++) {
        for(size_t i=0; i<4; i++)
            printf("%016llX  ", (unsigned long long) state[j][i]);
        printf("\n");
    }
#endif

    ap_uint<64> a0, a, b0, b;
    a0 = state[0][0] ^ state[1][1] ^ state[2][0] & state[3][0];
    b0 = state[0][1] ^ state[1][2] ^ state[2][1] & state[3][1];
    a = swap64(a0);
    b = swap64(b0);

    ap_uint<128> tag = ((ap_uint<128>) a << 64) | b;
    if (decrypt) {
        public_bus bdi;
        bdi_bus.read(bdi); 
        ap_uint<128> exp_tag = bdi.data >> 128;
        *msg_auth_valid = (tag == exp_tag) ? 1:0;
        #if DBG
            std::cout << "Reading tag ..." << std::endl;
        #endif
    } else {
        bdo.data = ((ap_uint<BLOCKSIZE>) tag) << 128;
        bdo_bus.write(bdo);
        #if DBG
            std::cout << "Writing tag ..." << std::endl;
        #endif
    }
    return;
}
